# BSD 2-Clause License

# Copyright 2022, Quoc-Tung Le

# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation 
#     and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.

import matplotlib.ticker

import torch
import torch.optim as optim

import numpy as np

import time

def training(m, support1, support2, lr = 1e-1, max_iter = 2000, option = 'GD', thresholding = 1e-10):
    # Iterative method: gradient descent, gradient descent momentum and ADAM
    # m: factorized matrix
    # support1: I in the paper
    # support2: J in the paper
    # lr: learning rate for iterative methods
    # max_iter: maximum number of iterations 
    # option: 'GD', 'GDMomentum' or 'Adam'
    # thresholding: if |m - XY| \leq threshold, then terminate

    # Calculate the number of elements in the support of X and Y
    init1 = np.sum(support1[:,0])
    init2 = np.sum(support2[:,0])

    support1 = torch.from_numpy(support1).double()
    support2 = torch.from_numpy(support2).double()

    # Initialization
    ft1 = (torch.rand(support1.shape[0], support1.shape[1]) / np.sqrt(init1)).double().requires_grad_()
    ft2 = (torch.rand(support2.shape[0], support2.shape[1]) / np.sqrt(init2)).double().requires_grad_()
    m = torch.from_numpy(m).double()

    # Choosing which algorithm to run
    if option == 'GD':
        optimizer = optim.SGD([ft1,ft2], lr = lr)
    elif option == 'GDMomentum':
        optimizer = optim.SGD([ft1,ft2], lr = lr, momentum=0.9)    
    else:
      optimizer = optim.Adam([ft1, ft2], lr = lr)
    
    loss_array = []
    time_array = []

    begin = time.time()
    for index in range(max_iter):
        optimizer.zero_grad()
        loss = 0.5 * torch.sum((torch.mm(ft1 * support1, ft2 * support2) - m) ** 2)
        loss.backward()
        optimizer.step()
        end = time.time()
        time_array.append(end - begin) 
        loss_array.append(torch.sqrt(2 * loss).data.numpy())
        if loss_array[-1] < thresholding:
            break
        if torch.isnan(loss):
            return False, loss_array, time_array
    
    return True, loss_array, time_array